import * as React from 'react';
import { observable, autorun, untracked } from 'mobx';
import { remove } from 'lodash';
import { EventProxy } from './eventProxy';

export const WidgetsContext = React.createContext({ parent: null });

function checkVisible({ _visible }) {
  return _visible !== false && _visible !== '';
}

function findWidgets(widget, parentType, filterFn, includeInvisibleDescendants) {
  if (!widget) return [];

  let { children = [] } = widget;
  let matched = children.filter(filterFn);

  // 过滤掉不可见后代
  if (!includeInvisibleDescendants) {
    children = children.filter((item) => checkVisible(item));
  }

  children.forEach((child) => {
    // 如果递归过程中发现了自己，则停止递归
    matched = matched.concat(findWidgets(child, parentType, filterFn, includeInvisibleDescendants));
  });

  // 过滤掉不可见的匹配项
  if (!includeInvisibleDescendants) {
    matched = matched.filter((item) => checkVisible(item));
  }

  return matched;
}

let retryQueue = [];
// 递归执行
export function retryDataBinds(tryTime = 10) {
  const fn = retryQueue.shift();
  if (!fn || tryTime <= 0) {
    retryQueue = [];
    return;
  }
  try {
    fn({ showLog: tryTime <= 1 });
  } catch (e) {
    console.error('retryDataBinds', e);
  }
  retryDataBinds(tryTime - 1);
}
export function createWidgets(widgetProps, dataBinds) {
  const nodeTree = createWidgetTree(widgetProps, dataBinds);
  const widgets = runFor(nodeTree, {}, null, null);
  return widgets;

  /**
   *
   * @param {*} curForNode
   * @param {*} forItems
   * @param {*} parentLevelWidgets
   * @param {*} parentWidget
   * @returns top level widgets or for dispose
   */
  function runFor(curForNode, forItems, parentLevelWidgets, parentWidget) {
    const nodeId = curForNode.id;
    if (!curForNode.value) {
      return createSubTree({ parentWidget, parentLevelWidgets }, curForNode, {});
    }
    const dispose = autorun(() => {
      let forList = [];
      try {
        forList = dataBinds[nodeId]._waFor(forItems);
      } catch (e) {
        console.error('waFor error', e);
      }

      if (!Array.isArray(forList)) {
        console.warn(nodeId, 'For 循环绑定的数据并不是数组，请检查');
        return;
      }

      // 让 forList 进行监听
      forList.forEach(() => {});
      untracked(() => {
        disposeWidgets(parentLevelWidgets[curForNode.id]);

        // clean nodes of previouse for run
        dfsTree(curForNode, (node) => {
          const arr = parentLevelWidgets[node.id];
          arr.splice(0, arr.length);
          if (parentWidget) {
            remove(parentWidget.children, ({ id }) => id === node.id);
          }
        });

        forList.forEach((item) => {
          createSubTree({ parentWidget, parentLevelWidgets }, curForNode, {
            ...forItems,
            [nodeId]: item,
          });
        });

        // 非初始化时遇到需要重新构建 dataBinds
        retryDataBinds();
      });
    });

    return dispose;
  }

  function createSubTree(ctx, curForNode, subForItems) {
    const widgets = { _disposers: [] };
    const { id: curForNodeId } = curForNode;
    const { parentWidget, parentLevelWidgets } = ctx;

    // traverse down the tree to set all widgets
    dfsTree(curForNode, (node, parentNode) => {
      if (node.forCount === curForNode.forCount) {
        // Leaf node
        const w = observable(widgetProps[node.id]);
        w.id = node.id;
        if (node === curForNode) {
          w._disposers = [];
        }
        widgets[node.id] = w;
        w._listeners = new EventProxy();
        // 提供一个给 Node 挂载 API 的方式
        untracked(() => {
          w.getConfig = () => ({}); // 兼容非复合组件调用 getConfig 能力
          w.findWidgets = (type, includeInvisibleDescendants) =>
            findWidgets(w, w.widgetType, type, includeInvisibleDescendants);
          w.getWidgetsByType = (type, includeInvisibleDescendants) =>
            w.findWidgets((currentWidget) => currentWidget.widgetType === type, includeInvisibleDescendants);
          w.extends = (name, fnOrData) =>
            Object.defineProperty(w, name, {
              value: fnOrData,
              writable: true,
            });
          w.on = (name, listener) => w._listeners.on(name, listener);
          w.off = (name, listener) => w._listeners.cancel(name, listener);
          w.getOwnerWidget = () => null; // 默认带上一个 widget
          // eslint-disable-next-line react/display-name
          const instanceRef = React.createRef();
          if (checkObserverInstanceRef(w.widgetType)) {
            w._instanceRef = instanceRef;
          } else {
            Object.defineProperty(w, '_instanceRef', {
              get() {
                return instanceRef;
              },
            });
          }
          w._getInstanceRef = () => w._instanceRef; // 默认初始值
          Object.defineProperty(w, '_methods', {
            get() {
              const instance = w._getInstanceRef();
              return instance?.current.methods;
            },
          });
        });
        w.children = [];
        const parent = parentNode ? widgets[parentNode.id] : parentWidget;
        if (parent) {
          w.parent = parent;
          // 只有可显示 visible 的才存入 children 里
          if (checkVisible(w)) {
            parent.children.push(w);
          }
        }
        parentLevelWidgets && parentLevelWidgets[node.id].push(w);

        // Setup data binds
        Object.keys(dataBinds[node.id] || {}).map((prop) => {
          if (prop === '_waFor') {
            return;
          }
          function getBindData(curForNodeId, options = {}) {
            let disposeError = false;
            const dispose = autorun((r) => {
              try {
                // Computed data bind in the next tick since data bind may read widgets data
                w[prop] = dataBinds[node.id][prop](subForItems);
                disposeError = false;
              } catch (e) {
                console.warn(`Error computing data bind ${w.id}.${prop}`, e);
                if (prop === '_waIf') {
                  w[prop] = false;
                } else {
                  options.showLog && console.warn(e);
                  retryQueue.push((opt) => {
                    r.dispose();
                    return getBindData(curForNodeId, opt);
                  });
                  disposeError = e;
                }
              }
            });

            if (!disposeError) {
              if (curForNodeId) {
                widgets[curForNodeId]._disposers.push(dispose);
              } else {
                widgets._disposers.push(dispose);
              }
            }
          }
          getBindData(curForNode.id);
        });
      } else {
        if (parentLevelWidgets) {
          const len = parentLevelWidgets[node.id].push([]);
          widgets[node.id] = parentLevelWidgets[node.id][len - 1];
        } else {
          widgets[node.id] = observable([]);
        }
      }
    });

    // run for of next level
    dfsTree(curForNode, (node, parentNode) => {
      if (node.forCount === curForNode.forCount + 1 && dataBinds[node.id]?._waFor) {
        widgets[node.id]._disposers = { dataBinds: [] };
        const dispose = runFor(node, subForItems, widgets, node.parent && widgets[node.parent.id]);
        if (curForNodeId) {
          widgets[curForNodeId]._disposers.push(dispose);
        }
      }
    });

    return widgets;
  }
}

/**
 * Add parent, children to widget
 */
function createWidgetTree(widgets, dataBinds) {
  const virtualRoot = { children: [], forCount: 0 };
  const nodes = Object.keys(widgets).reduce((result, id) => {
    result[id] = {
      id,
      value: widgets[id],
      children: [],
      parent: null,
      forCount: 0,
    };
    return result;
  }, {});

  // Create widgets tree API
  Object.keys(nodes).map((id) => {
    const curNode = nodes[id];
    const parent = nodes[widgets[id]._parentId];
    if (!parent) {
      virtualRoot.children.push(curNode);
      return;
    }
    curNode.parent = parent;
    parent.children.push(curNode);
  });

  virtualRoot.children.map(addForCount);

  // dfs, add forCount
  function addForCount(node) {
    if (node.parent) {
      node.forCount = node.parent.forCount;
    }
    if (dataBinds[node.id] && dataBinds[node.id]._waFor) {
      node.forCount += 1;
    }
    node.children.map(addForCount);
  }

  return virtualRoot;
}

function dfsTree(node, fn, parent) {
  node.value && fn(node, parent);
  node.children.map((e) => dfsTree(e, fn, node.value ? node : null));
}

// dispose autorun
function disposeWidgets(widgets = []) {
  widgets.forEach((widget) => {
    const disposers = widget._disposers;
    if (disposers) {
      disposers.map((dispose) => dispose());
      disposers.splice(0, disposers.length);
    }
    disposeWidgets(widget.children);
  });
}

function checkObserverInstanceRef(sourceKey = '') {
  return !sourceKey?.startsWith('CLOUDBASE_BUSSINESS') /*&& ENABLE_OBSERVER_REF*/;
}
